using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Drawing.Drawing2D;
using System.Diagnostics;

namespace Transparency
{
    /// <summary>
    /// Summary description for TransparentForm.
    /// </summary>
    public class TransparentForm : System.Windows.Forms.Form
    {
        private IContainer components;
        internal const int WM_NCHITTEST = 0x84; // winuser.h
        internal const int WM_SETCURSOR = 0x0020;
        internal const int WM_MOUSEACTIVATE = 0x0021;
        internal const int WM_WINDOWPOSCHANGED = 0x0047;
        internal const int HTCLIENT = 1;
        internal const int HTCAPTION = 2;
        internal const int HTLEFT = 10;
        internal const int HTRIGHT = 11;
        internal const int HTTOP = 12;
        internal const int HTTOPLEFT = 13;
        internal const int HTTOPRIGHT = 14;
        internal const int HTBOTTOM = 15;
        internal const int HTBOTTOMLEFT = 16;
        internal const int HTBOTTOMRIGHT = 17;
        int penWidth = 4;

        public TransparentForm()
        {
            //
            // Required for Windows Form Designer support
            //
            InitializeComponent();

            this.KeyPreview = true;
            this.FormBorderStyle = FormBorderStyle.None;

        }

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (components != null)
                {
                    components.Dispose();
                }
            }
            base.Dispose(disposing);
        }

        #region Windows Form Designer generated code
        /// <summary>
        /// Required method for Designer support - do not modify
        /// the contents of this method with the code editor.
        /// </summary>
        private void InitializeComponent()
        {
            this.checkBoxWndProc = new System.Windows.Forms.CheckBox();
            this.SuspendLayout();
            // 
            // checkBoxWndProc
            // 
            this.checkBoxWndProc.Anchor = System.Windows.Forms.AnchorStyles.Top;
            this.checkBoxWndProc.AutoSize = true;
            this.checkBoxWndProc.BackColor = System.Drawing.Color.Transparent;
            this.checkBoxWndProc.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
            this.checkBoxWndProc.Location = new System.Drawing.Point(157, 43);
            this.checkBoxWndProc.Name = "checkBoxWndProc";
            this.checkBoxWndProc.Size = new System.Drawing.Size(179, 30);
            this.checkBoxWndProc.TabIndex = 0;
            this.checkBoxWndProc.Text = "Use WndProc";
            this.checkBoxWndProc.UseVisualStyleBackColor = false;
            this.checkBoxWndProc.KeyDown += new System.Windows.Forms.KeyEventHandler(this.checkBox1_KeyDown);
            this.checkBoxWndProc.KeyPress += new System.Windows.Forms.KeyPressEventHandler(this.checkBox1_KeyPress);
            // 
            // TransparentForm
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(15, 37);
            this.ClientSize = new System.Drawing.Size(516, 372);
            this.Controls.Add(this.checkBoxWndProc);
            this.Font = new System.Drawing.Font("Microsoft Sans Serif", 12F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
            this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.None;
            this.Name = "TransparentForm";
            this.Text = "Transparency Test";
            this.TransparencyKey = System.Drawing.Color.White;
            this.Load += new System.EventHandler(this.TransparentForm_Load);
            this.SizeChanged += new System.EventHandler(this.TransparentForm_SizeChanged);
            this.Paint += new System.Windows.Forms.PaintEventHandler(this.TransparentForm_Paint);
            this.DoubleClick += new System.EventHandler(this.TransparentForm_DoubleClick);
            this.KeyDown += new System.Windows.Forms.KeyEventHandler(this.TransparentForm_KeyDown);
            this.KeyPress += new System.Windows.Forms.KeyPressEventHandler(this.TransparentForm_KeyPress);
            this.MouseDown += new System.Windows.Forms.MouseEventHandler(this.TransparentForm_MouseDown);
            this.MouseMove += new System.Windows.Forms.MouseEventHandler(this.TransparentForm_MouseMove);
            this.MouseUp += new System.Windows.Forms.MouseEventHandler(this.TransparentForm_MouseUp);
            this.ResumeLayout(false);
            this.PerformLayout();

        }
        #endregion


        void TransparentForm_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            using (Pen pen = new Pen(Color.DarkBlue, penWidth))
            {
                Rectangle rect = this.ClientRectangle;
                rect.Inflate(-penWidth / 2, -penWidth / 2);
                g.FillEllipse(Brushes.Yellow, rect);
                g.DrawEllipse(pen, rect);
            }
            using (Font font = new Font(this.Font, FontStyle.Bold))
            {
                StringFormat format = new StringFormat();
                format.Alignment = StringAlignment.Center;
                format.LineAlignment = StringAlignment.Center;
                g.DrawString("an ellipse", font, Brushes.DarkBlue, this.ClientRectangle, format);
            }
        }


        GraphicsPath pathInner;
        void SetEllipseRegion()
        {
            Rectangle rect = this.ClientRectangle;
            // Assume: this.FormBorderStyle = FormBorderStyle.None
            using (GraphicsPath path = new GraphicsPath())
            {
                path.AddEllipse(rect);
                this.Region = new Region(path);
            }
            pathInner = new GraphicsPath();
            rect.Inflate(-penWidth / 2, -penWidth / 2);
            pathInner.AddEllipse(rect);
        }

        void TransparentForm_Load(object sender, EventArgs e)
        {
            SetEllipseRegion();
        }

        void TransparentForm_SizeChanged(object sender, EventArgs e)
        {
            SetEllipseRegion();
        }

        private CheckBox checkBoxWndProc;

        Point downPoint = Point.Empty;

        void TransparentForm_MouseDown(object sender, MouseEventArgs e)
        {
            if (checkBoxWndProc.Checked) return;
            if (e.Button != MouseButtons.Left) return;
            downPoint = new Point(e.X, e.Y);
        }

        void TransparentForm_MouseMove(object sender, MouseEventArgs e)
        {
            if (checkBoxWndProc.Checked) return;
            if (downPoint == Point.Empty) return;
            Point location =
              new Point(
              this.Left + e.X - downPoint.X, this.Top + e.Y - downPoint.Y);
            Console.WriteLine("({0},{1}),({2},{3}),({4},{5})",
                e.X, e.Y, downPoint.X, downPoint.Y, this.Left, this.Top);
            this.Location = location;
        }

        void TransparentForm_MouseUp(object sender, MouseEventArgs e)
        {
            if (checkBoxWndProc.Checked) return;
            if (e.Button != MouseButtons.Left) return;
            downPoint = Point.Empty;
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            Point location = new Point(this.Left, this.Top);
            //Handle arrow keys
            switch (keyData)
            {
                case Keys.Up:
                    --location.Y;
                    break;

                case Keys.Left:
                    --location.X;
                    break;

                case Keys.Down:
                    ++location.Y;
                    break;

                case Keys.Right:
                    ++location.X;
                    break;

                default:
                    //other command keys are done in base
                    return base.ProcessCmdKey(ref msg, keyData);
            }

            //For arrow keys, change location and indicate handled
            this.Location = location;
            return true;
        }

        void TransparentForm_KeyDown(object sender, KeyEventArgs e)
        {

            Point location = new Point(this.Left, this.Top);

            switch (e.KeyCode)
            {
                case Keys.I:
                case Keys.Up:
                    --location.Y;
                    break;

                case Keys.J:
                case Keys.Left:
                    --location.X;
                    break;

                case Keys.K:
                case Keys.Down:
                    ++location.Y;
                    break;

                case Keys.L:
                case Keys.Right:
                    ++location.X;
                    break;
            }

            this.Location = location;
        }

        void TransparentForm_KeyPress(object sender, KeyPressEventArgs e)
        {
            System.Console.WriteLine("Press");

            Point location = new Point(this.Left, this.Top);

            switch (e.KeyChar)
            {
                case 'i':
                    --location.Y;
                    break;

                case 'j':
                    --location.X;
                    break;

                case 'k':
                    ++location.Y;
                    break;

                case 'l':
                    ++location.X;
                    break;
            }

            this.Location = location;
        }

        void TransparentForm_KeyPress_Save(object sender, KeyPressEventArgs e)
        {
            Point location = new Point(this.Left, this.Top);

            switch (e.KeyChar)
            {
                case 'i':
                    --location.Y;
                    break;

                case 'j':
                    --location.X;
                    break;

                case 'k':
                    ++location.Y;
                    break;

                case 'l':
                    ++location.X;
                    break;
            }

            this.Location = location;
        }

        private void TransparentForm_DoubleClick(object sender, System.EventArgs e)
        {
            DialogResult res = MessageBox.Show("Close window?",
                "Close Window", MessageBoxButtons.YesNo);
            if (res == DialogResult.Yes)
            {
                this.Close();
            }
        }

        private void WriteDebug(String msg)
        {
            if (bDebug)
            {
                Debug.WriteLine(msg);
            }
        }

        bool bDebug = true;
        private void AlterMessage(ref Message m)
        {
            if (checkBoxWndProc.Checked)
            {
                WriteDebug(String.Format("---{0}\nIn : {1}", ++count, m.ToString()));

                if (m.Msg == WM_WINDOWPOSCHANGED)
                {
                    this.Invalidate();
                }

                // If the user clicked on the client area,
                // ask the OS to treat it as a click on the caption
                if (m.Result.ToInt32() == HTCLIENT)
                {
                    m.Result = (IntPtr)HTCAPTION;
                }

                #region Process Edge Locations for cursor and resize
                //Process edge locations

                if (m.Msg != WM_SETCURSOR && m.Msg != WM_NCHITTEST && m.Msg != WM_MOUSEACTIVATE)
                {
                    return;
                }

                // Map point around center of ellipse
                Point clientLocation = this.PointToClient(new Point(m.LParam.ToInt32()));
                // Only react to points close to edge
                if (pathInner.IsVisible(clientLocation)) return;
                WriteDebug("Edge: " + m.ToString());
                
                //Locate relative coordinates
                int x = clientLocation.X - ClientRectangle.Width / 2;
                int y = -(clientLocation.Y - ClientRectangle.Height / 2);
                int absX = Math.Abs(x);
                int absY = Math.Abs(y);

                //Leave room for center
                int centerEdge = penWidth;
                // Check if x or y is close to edge. This works for points close to bounding
                // rectangle, but does not work for rounded portion of ellipse
                if (absX < centerEdge && y < 0) { m.Result = (IntPtr)HTBOTTOM; return; }
                else if (absX < centerEdge && y > 0) { m.Result = (IntPtr)HTTOP; return; }
                else if (absY < centerEdge && x > 0) { m.Result = (IntPtr)HTRIGHT; return; }
                else if (absY < centerEdge && x < 0) { m.Result = (IntPtr)HTLEFT; return; }
                else Debug.Assert(x != 0);

                float slope = (float)y / (float)x;
                if (x > 0 && y > 0)
                {
                    // Q1
                    if (slope > 1.67) m.Result = (IntPtr)HTTOP;
                    else if (slope < .5) m.Result = (IntPtr)HTRIGHT;
                    else m.Result = (IntPtr)HTTOPRIGHT;
                }
                else if (x > 0 && y < 0)
                {
                    // Q4
                    if (slope < -1.67) m.Result = (IntPtr)HTBOTTOM;
                    else if (slope > -0.5) m.Result = (IntPtr)HTRIGHT;
                    else m.Result = (IntPtr)HTBOTTOMRIGHT;
                }
                else if (x < 0 && y < 0)
                {
                    // Q3
                    if (slope > 1.67) m.Result = (IntPtr)HTBOTTOM;
                    else if (slope < .5) m.Result = (IntPtr)HTLEFT;
                    else m.Result = (IntPtr)HTBOTTOMLEFT;
                }
                else
                {
                    // Q2
                    if (slope < -1.67) m.Result = (IntPtr)HTTOP;
                    else if (slope > -0.5) m.Result = (IntPtr)HTLEFT;
                    else m.Result = (IntPtr)HTTOPLEFT;
                }
                #endregion
            }
        }
        static int count = 0;
        protected override void WndProc(ref Message m)
        {
           
            // Let the base class have first crack
            base.WndProc(ref m);
            AlterMessage(ref m);
        }

        private void checkBox1_KeyDown(object sender, KeyEventArgs e)
        {
            System.Console.WriteLine("Button Down");
        }

        private void checkBox1_KeyPress(object sender, KeyPressEventArgs e)
        {
            System.Console.WriteLine("Button Press");
        }

    }
}